package org.rapidoid;

/*
 * #%L
 * rapidoid-net
 * %%
 * Copyright (C) 2014 - 2015 Nikolche Mihajlovski
 * %%
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * #L%
 */

import org.rapidoid.annotation.Authors;
import org.rapidoid.annotation.Since;
import org.rapidoid.log.Log;
import org.rapidoid.log.LogLevel;
import org.rapidoid.net.Protocol;
import org.rapidoid.net.TCP;
import org.rapidoid.net.TCPClient;
import org.rapidoid.net.TCPServer;
import org.rapidoid.net.abstracts.Channel;
import org.rapidoid.util.UTILS;
import org.testng.annotations.Test;

@Authors("Nikolche Mihajlovski")
@Since("2.0.0")
public class TcpClientTest extends NetTestCommons {

	private final Protocol UPPERCASE_SERVER = new Protocol() {
		@Override
		public void process(Channel ctx) {
			if (ctx.isInitial()) {
				return;
			}
			ctx.writeln(ctx.readln().toUpperCase());
		}
	};

	private final Protocol HI_CLIENT = new Protocol() {
		@Override
		public void process(Channel ctx) {
			if (ctx.isInitial()) {
				ctx.writeln("Hi there!");
				return;
			}

			eq(ctx.readln(), "HI THERE!");
		}
	};

	@Test
	public void testTCPClientWithDefaultConnections() {
		Log.setLogLevel(LogLevel.DEBUG);

		TCPClient client = TCP.client().host("localhost").port(8080).connections(5).protocol(HI_CLIENT).build().start();

		// let the clients wait
		UTILS.sleep(3000);

		TCPServer server = TCP.server().port(8080).protocol(UPPERCASE_SERVER).build().start();

		// let the server serve the clients
		UTILS.sleep(3000);

		eq(client.info().messagesProcessed(), 5);
		eq(server.info().messagesProcessed(), 5);

		client.shutdown();
		server.shutdown();
	}

	@Test
	public void testTCPClientWithCustomConnections() {
		Log.setLogLevel(LogLevel.DEBUG);

		TCPClient client = TCP.client().build().start();
		client.connect("localhost", 8080, HI_CLIENT, 10);

		// let the clients wait
		UTILS.sleep(3000);

		TCPServer server = TCP.server().port(8080).protocol(UPPERCASE_SERVER).build().start();

		// let the server serve the clients
		UTILS.sleep(3000);

		eq(client.info().messagesProcessed(), 10);
		eq(server.info().messagesProcessed(), 10);

		client.shutdown();
		server.shutdown();
	}

	@Test
	public void testTCPClientWithDefaultAndCustomConnections() {
		Log.setLogLevel(LogLevel.DEBUG);

		TCPClient client = TCP.client().host("127.0.0.1").port(8080).connections(3).protocol(HI_CLIENT).build().start();
		client.connect("localhost", 9090, HI_CLIENT, 2);

		// let the clients wait
		UTILS.sleep(3000);

		TCPServer server1 = TCP.server().port(8080).protocol(UPPERCASE_SERVER).build().start();
		TCPServer server2 = TCP.server().port(9090).protocol(UPPERCASE_SERVER).build().start();

		// let the servers serve the clients
		UTILS.sleep(3000);

		eq(client.info().messagesProcessed(), 5);
		eq(server1.info().messagesProcessed(), 3);
		eq(server2.info().messagesProcessed(), 2);

		client.shutdown();
		server1.shutdown();
		server2.shutdown();
	}

}
